#!/usr/bin/env python

# Read pcap file, anonymize MAC and IP addresses and write to new pcap file

# MAC addresses are set to all zeros.
# IP addresses are mapped to new ones IF they match any of the defined anonymize_networks.

# IPv4 addresses are rewritten to 240.xx.yy.zz (with offset = 240)
# IPv6 addresses are rewritten to 8683::F0xx:yyzz (with offset = 240 = 0xF0)

# IP checksum is rewritten after packet alteration to correct value.
# Neither UDP/TCP headers nor payload is rewritten.

### There are two mapping algorithms implemented:

# 1. Keyed-hashing
# If you pass a --secret 'foobar' on command line, IP addresses will be mapped to 240.xx.yy.zz,
# where xxyyzz are the first 24 bit of the keyed-hash value (HMAC).
# The mapping is deterministic and can be reversed with knowledge of the secret.

# 2. First time seen
# If you do not set a secret, first IP address seen is mapped to 240.0.0.1, second to 240.0.0.2 etc.
# The mapping is unique to the order of seen IP addresses from the input file, it is NOT saved.

# 2013 Matthaeus Wander <matthaeus.wander@uni-due.de>

import argparse
import dpkt
import hashlib
import hmac
import networklist
import signal
import socket
import struct
import sys

# config

anonymize_networks = networklist.networklist([
    '134.91.0.0/16',
    '132.252.0.0/16',
    '2001:638:501::/48',
])

ignore_ethertypes = [
    dpkt.ethernet.ETH_TYPE_ARP,
]

# end of config

ipmap = {}
ipcount = 0

def ip_mapping(binip):
    # keyed-hash
    if args.secret and len(args.secret) > 0:
        hmac_obj = hmac.new(args.secret, binip, hashlib.sha1)
        ipv4 = struct.pack("!B", args.offset) + hmac_obj.digest()[0:3]
    # first time seen
    else:
        newnum = len(ipmap) + 1
        ipv4 = struct.pack("!I", args.offset<<24 | newnum)

    if len(binip) == 4:
        return ipv4
    elif len(binip) == 16:
        return '\x86\x83\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' + ipv4
    else:
        raise Exception("len(n_ip) is neither 4 nor 16:", len(binip))

def replace_address(binip):
    if binip not in ipmap:
        ipmap[binip] = ip_mapping(binip)

    return ipmap[binip]

parser = argparse.ArgumentParser(description='anonymize pcap file')
parser.add_argument('-o', '--offset', help='one byte offset for IP address range (240-254 are reasonable choices, default 240)', default=240, type=int)
parser.add_argument('-s', '--secret', help='set secret for keyed-hash mapping (default none)')
parser.add_argument('file_in', help='input pcap filename. If "-", stdin is used')
parser.add_argument('file_out', help='output pcap filename. If "-", stdout is used')
args = parser.parse_args()

# ignore ctrl+c (will terminate when EOF has been read)
signal.signal(signal.SIGINT, signal.SIG_IGN)

infile  = sys.stdin  if args.file_in  is '-' else file(args.file_in, "rb")
outfile = sys.stdout if args.file_out is '-' else file(args.file_out, "wb")
if not 0 < args.offset < 255:
    raise Exception("offset must be 0-255, not " + str(args.offset))

if args.secret and len(args.secret) > 0:
    print >>sys.stderr, "Got secret, using keyed-hash mapping"
else:
    print >>sys.stderr, "No secret given, using first time seen mapping"

pcapReader = dpkt.pcap.Reader(infile)
linktype = pcapReader.datalink()
pcapWriter = dpkt.pcap.Writer(outfile, linktype=linktype)

for ts, data in pcapReader:
    if linktype == dpkt.pcap.DLT_EN10MB: # Ethernet
        frame = dpkt.ethernet.Ethernet(data)
        # clear MAC addresses
        frame.src = '\x00' * len(frame.src)
        frame.dst = '\x00' * len(frame.dst)
        packettype = frame.type
    elif linktype == dpkt.pcap.DLT_LINUX_SLL: # Linux cooked capture (SLL)
        frame = dpkt.sll.SLL(data)
        # clear MAC address
        frame.hdr = '\x00' * len(frame.hdr)
        packettype = frame.ethtype
    else:
        raise Exception("Linktype " + str(linktype) + " is not implemented yet. Please report this error.")

    if packettype not in (dpkt.ethernet.ETH_TYPE_IP, dpkt.ethernet.ETH_TYPE_IP6):
        if packettype not in ignore_ethertypes:
            print >> sys.stderr, "Warning: skipping unknown EtherType", hex(packettype) + ". Are you using this network protocol or did the parser fail to read ethernet frames?", packettype
            pcapWriter.writepkt(frame, ts)
        continue

    # IPv4/IPv6
    ip = frame.data
    if packettype == dpkt.ethernet.ETH_TYPE_IP:
        p_src = socket.inet_ntop(socket.AF_INET, ip.src)
        p_dst = socket.inet_ntop(socket.AF_INET, ip.dst)
    else:
        p_src = socket.inet_ntop(socket.AF_INET6, ip.src)
        p_dst = socket.inet_ntop(socket.AF_INET6, ip.dst)

    if p_src in anonymize_networks:
        ip.src = replace_address(ip.src)
        ip.sum = 0 # remove cksum, dpkt serializer will compute new value
    if p_dst in anonymize_networks:
        ip.dst = replace_address(ip.dst)
        ip.sum = 0 # remove cksum, dpkt serializer will compute new value
    
    pcapWriter.writepkt(frame, ts)

pcapWriter.close()
print >>sys.stderr, "Finished. Used", len(ipmap), "IP address mappings"
